/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2004-2005, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL), as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the LGPL in the file COPYING for more
 * details.
 *
 */
                                                                              
using System.Runtime.InteropServices;
using System;
using System.Text; // for StringBuilder

namespace Fuse
{
    /** Mode flags for some common types.  Used for testing / checking
     * Stat.mode value.
    */
    public enum ModeFlags
    {
	Socket       = 49152, //< octal 0140000
	SymLink      = 40960, //< octal 0120000
	RegularFile  = 32768, //< octal 0100000
	BlockDevice  = 24576, //< octal 0060000
	Directory    = 16384, //< octal 0040000
	CharDevice   = 8192,  //< octal 0020000
	FIFO         = 4096,  //< octal 0010000
	PermissionMask = 511, //< octal 0000777, all file permissions
	ReadWriteAll = 438,   //< octal 0000666, read+write permission for all
	ReadAll      = 292    //< octal 0000444, read permission for all
    }

    // XXX: Warning - this must be in sync with wrapper.{c,h} code
    /** @struct Stat
      @brief Used for returning information about a file or directory.

      @warning This must match the definition in wrapper.h, as it is used in
      P/Invoke calls.
    */
    public struct Stat
    {
	public ulong ino; // inode number
	public uint mode;  // permissions bits
	public uint nlink; // number of hard links
	public uint uid;   // user id of owner
	public uint gid;   // group id of owner
	public uint rdev;  // device type (if inode device)
	public ulong size; // size in bytes
	public long blocks; // number of blocks
	public ulong atime; // time of last access
	public ulong mtime; // time of last data modification
	public ulong ctime; // time of last metadata change
	public uint atimensec;
	public uint mtimensec;
	public uint ctimensec;

	/** Set time fields based on a System.DateTime.
	  @param time Time to set, in UTC

	  @warning If the time specified is not in UTC, then the resulting time
	  may be offset.  
	  
	  To set the time as 'now':
	  @code
	  Stat stat = new Stat();
	  stat.SetTimes( DateTime.UtcNow );
	  @endcode
	*/
	public void SetTimes( System.DateTime time )
	{
	    TimeSpan t = time - new DateTime(1970,1,1);

	    // diff is in 100 nanosecond increments
	    long time_sec = (long)t.TotalSeconds;
	    long time_nsec = 100*(t.Ticks - time_sec*TimeSpan.TicksPerSecond);
	    mtime = ctime = atime = (ulong)time_sec;
	    mtimensec = ctimensec = atimensec = (uint)time_nsec;
	}

	public fuse_attr ToFuseAttr()
	{
	    fuse_attr attr = new fuse_attr();
	    attr.ino = ino;
	    attr.mode = mode;
	    attr.nlink = nlink;
	    attr.uid = uid;
	    attr.gid = gid;
	    attr.rdev = rdev;
	    attr.size = size;
	    attr.blocks = checked((ulong)blocks);
	    attr.atime = atime;
	    attr.mtime = mtime;
	    attr.ctime = ctime;
	    attr.atimensec = atimensec;
	    attr.mtimensec = mtimensec;
	    attr.ctimensec = ctimensec;
	    return attr;
	}

	public static Fuse.Stat ToStat( fuse_attr attr )
	{
	    Fuse.Stat stat = new Fuse.Stat();
	    stat.ino = attr.ino;
	    stat.mode = attr.mode;
	    stat.nlink = attr.nlink;
	    stat.uid = attr.uid;
	    stat.gid = attr.gid;
	    stat.rdev = attr.rdev;
	    stat.size = attr.size;
	    stat.blocks = checked((long)attr.blocks);
	    stat.atime = attr.atime;
	    stat.mtime = attr.mtime;
	    stat.ctime = attr.ctime;
	    stat.atimensec = attr.atimensec;
	    stat.mtimensec = attr.mtimensec;
	    stat.ctimensec = attr.ctimensec;
	    return stat;
	}
    }

    /** @struct StatFS
      @brief Structure for returning information about a filesystem.

      @warning This must match the definition in wrapper.h, as it is
      used in P/Invoke calls.
    */
    public struct StatFS
    {
	public int blockSize;
	public long blocks;
	public long blocksFree;
	public long blocksAvail;
	public long files;
	public long filesFree;
	public int namelen;

	/** Convert to a fuse_kstatfs struct
	*/
	public fuse_kstatfs ToKStatFS()
	{
	    fuse_kstatfs kstat = new fuse_kstatfs();
	    kstat.bsize   = checked((uint)blockSize);
	    kstat.blocks  = checked((ulong)blocks);
	    kstat.bfree   = checked((ulong)blocksFree);
	    kstat.bavail  = checked((ulong)blocksAvail);
	    kstat.files   = checked((ulong)files);
	    kstat.ffree   = checked((ulong)filesFree);
	    kstat.namelen = checked((uint)namelen);
	    return kstat;
	}
    }

    /** @class Dirent
      @brief result from Wrapper.readdir() call.

      The system dirent struct is a variable length struct containing a string
      at the end, which isn't trivial to marshal.  So the readdir method
      converts to Dirent.
    */
    public class Dirent
    {
	public ulong  ino;
	public long   off;
	public ushort reclen;
	public byte   type;
	public string name;
    }

    /** @class Wrapper
      @brief static methods which wrap around system calls

      Some of these could be handled using Mono.Posix (or Mono.Unix?), but I'm
      trying to remove dependency on the Mono runtime.

      The reason that file descriptors are 64 bit is because that is how they
      are used in FUSE.  That way they should work just as well on a 64bit
      system as a 32bit system (perhaps they should really be an IntPtr).
    */
    public class Wrapper
    {
	internal const string RawFSLibName = "FuseWrapper";
	internal const string LibcName = "libc";
	internal const int NAME_MAX = 256; // XXX 


	/** System chmod call wrapper
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_chmod")] 
	public extern static int chmod( string name, uint mode );

	/** System chown call wrapper
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_chown")] 
	public extern static int chown( string name, int uid, int gid );

	/** System closedir wrapper
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_closedir")] 
	public extern static int closedir( IntPtr handle );

	/** System close wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_close")] 
	public extern static int close( IntPtr fd );

	/** System fstat wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_fstat")] 
	public extern static int fstat( IntPtr fd, ref Stat stat );

	/** System link wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_link")] 
	public extern static int link( string oldpath, string newpath );

	/** System lstat wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_lstat")] 
	public extern static int lstat( string name, ref Stat stat );

	/** System mkdir wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_mkdir")] 
	public extern static int mkdir( string name, uint mode );

	/** System mknod wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_mknod")] 
	public extern static int mknod( string name, uint mode, uint rdev );

	/** System opendir wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_opendir")] 
	public extern static IntPtr opendir( string name, out int err );

	/** System open wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_open")] 
	public extern static IntPtr open( string name, int flags, out int err );

	/** Wrapper for system readlink call.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_readlink")] 
	public extern static int readlink( string name, 
		    [Out] StringBuilder buf, int bufLen );

	/** Simplified wrapper for readlink call.
	    @returns link name as a string, or null if the file is not a link
	*/
	public static string readlink( string name )
	{
	    StringBuilder buf = new StringBuilder();
	    buf.Capacity = NAME_MAX;
	    int res = readlink( name, buf, (int)buf.Capacity );
	    if(res > 0)
		return buf.ToString();
	    else
		return null;
	}

	[DllImport (RawFSLibName, EntryPoint="wrapper_readdir")] 
	private extern static int sys_readdir( IntPtr handle,
		out _Dirent dentry, [Out] StringBuilder buf, int bufLen );
   
	/** Readdir wrapper.
	    @returns Dirent entry for the next item.
	*/
	public static Dirent readdir ( IntPtr handle )
	{
	    _Dirent dentry;
	    StringBuilder buf = new StringBuilder();
	    buf.Capacity = NAME_MAX;
	    int res = sys_readdir(handle, out dentry, buf, (int)buf.Capacity);
	    if(res < 0)
		return null;
	    Dirent d = new Dirent();
	    CopyDirent(d, ref dentry);
	    d.name = buf.ToString();
	    return d;
	}

	/** System rename wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_rename")] 
	public extern static int rename( string oldpath, string newpath );

	/** System rmdir wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_rmdir")] 
	public extern static int rmdir( string path );

	/** System statfs wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_statfs")] 
	public extern static int statfs( string name, ref StatFS stat );

	/** System symlink wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_symlink")] 
	public extern static int symlink( string oldpath, string newpath );

	/** System sync wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_sync")] 
	public extern static int sync( IntPtr fd, uint datasync );

	/** System truncate wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_truncate")] 
	public extern static int truncate( string name, ulong size );

	/** System unlink wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_unlink")] 
	public extern static int unlink( string path );

	/** System utime wrapper.
	    @returns 0 on success, or a negative error number on failure.
	*/
	[DllImport (RawFSLibName, EntryPoint="wrapper_utime")] 
	public extern static int utime( string name, ulong atime, ulong mtime);


	//[DllImport (RawFSLibName, EntryPoint="wrapper_readdir")]
	//public extern static string readdir( IntPtr handle );
	

	// Must match definition of FuseDirent in wrapper C code
	private struct _Dirent
	{
	    public ulong   d_ino;
	    public long    d_off;
	    public ushort  d_reclen;
	    public byte    d_type;
	}

	private static void CopyDirent( Dirent to, ref _Dirent from )
	{
	    to.ino = from.d_ino;
	    to.off = from.d_off;
	    to.reclen = from.d_reclen;
	    to.type = from.d_type;
	}

    }

}
